(***********************************************************************)
(*                              Project Couverture                     *)
(*                                                                     *)
(* file: htmlReport.ml                                                 *)
(* authors: Adrien Jonquet, Philippe Wang, Alexis Darrasse             *)
(* licence: CeCIL-B                                                    *)
(***********************************************************************)

type html = 
    | A of attributes * children
    | SPAN of attributes * children
    | DIV of attributes * children
    | BR
    | DATA of string
and attributes = (string * string) list 
and tagname = string
and children = html list

type sort = 
    | Statement of id
    | Condition of id
    | Decision  of id
and covered_or_not = Covered | Notcovered
and id = int

type decorated_char = 
    { 
      position: int*int;
      data : string;
      parents : segment list
    }
and segment = 
    {
      start_position : int;
      end_position : int;
      expression_sort : sort;
      coverage_status : covered_or_not;
    }

let htmlspecialchars (s : string) : string =
  let buffer = Buffer.create (String.length s * 2) in
    for i = 0 to String.length s - 1 do
      match s.[i] with
        | '<' -> Buffer.add_string buffer "&lt;"
        | '>' -> Buffer.add_string buffer  "&gt;"
        | '&' -> Buffer.add_string buffer  "&amp;"
        | '\'' -> Buffer.add_string buffer "&apos;"
        | '"' -> Buffer.add_string buffer "&quot;"
        | c -> Buffer.add_char buffer c
    done;
    Buffer.contents buffer

let string_of_html html =
  let b = Buffer.create 2048 in
  let rec convert = function
    | A (attributes, children) ->
        Buffer.add_string b "<A";
        convert_attributes b attributes;
        Buffer.add_string b ">";
        List.iter convert children;
        Buffer.add_string b "</A>"
    | BR -> Buffer.add_string b "<BR/>"
    | DIV (attributes, children) ->
        Buffer.add_string b "<DIV";
        convert_attributes b attributes;
        Buffer.add_string b ">";
        List.iter convert children;
        Buffer.add_string b "</DIV>"
    | SPAN (attributes, children) ->
        Buffer.add_string b "<SPAN";
        convert_attributes b attributes;
        Buffer.add_string b ">";
        List.iter convert children;
        Buffer.add_string b "</SPAN>"
    | DATA s -> Buffer.add_string b (htmlspecialchars s)
  and convert_attributes b attributes =
    List.iter 
      (fun (a,v) -> Printf.bprintf b " %s='%s'" a (htmlspecialchars v)) 
      attributes
  in convert html;
    Buffer.contents b

let head co title =
  output_string co
    ("<HEAD>
      <TITLE>"^title^"</TITLE>
      <LINK rel=\"stylesheet\" type=\"text/css\" href=\"coverage.css\">
      <META http-equiv='Content-Type' content='text/html; charset=ISO-8859-1'>
      <SCRIPT TYPE=text/javascript>
        function show(x) {
          x.style.fontStyle = 'italic';
          x.style.fontWeight = 'bold';
        }
        function hide(x) {
          x.style.fontStyle = 'normal';
          x.style.fontWeight = 'normal';
        }

      </SCRIPT>
  </HEAD>")

let head_object co =
  output_string co
    "
<HEAD>
  <TITLE>Object code coverage report</TITLE>
  <LINK rel=\"stylesheet\" type=\"text/css\" href=\"coverage.css\" />
  <META http-equiv='Content-Type' content='text/html; charset=ISO-8859-1' />
</HEAD>"


let generate_css co =
  output_string co
    "
* {
  margin: 0;
  padding: 0;
}

body { margin-top:1em; line-height: 150%; font-family:'Damascus', 'Trebuchet', 'Helvetica', 'sans-serif'; }


h1, h2, h3, h4, h5 { padding:10px 20px 10px 20px; }

br, hr { margin:1ex; padding:1ex; }

a, a:hover { font: inherit; text-decoration:none ; color:inherit; }

pre {border-left-style:solid;border-left-color:#222;border-left-width:1ex;padding-left:1ex;}

.p-covered_decision       {padding:2px 0px 2px 0px;border-width:2px;border-bottom-style:solid;border-bottom-color:green}
.p-not_covered_decision   {padding:4px 0px 4px 0px;border-width:2px;border-bottom-style:solid;border-bottom-color:red}
.p-covered_condition      {padding:2px 0px 2px 0px;border-width:2px;border-top-style:solid;border-top-color:green}
.p-not_covered_condition  {padding:4px 0px 4px 0px;border-width:2px;border-top-style:solid;border-top-color:red}
.p-covered_statement      {padding:2px 0px 2px 0px;background-color:#CCFF66}
.p-not_covered_statement  {padding:4px 0px 4px 0px;background-color:#FF6666}


.bytecode  {font-family:monospace}

.covered { background-color: lightgreen ; padding:0.7ex; margin:0.3ex; }
.uncovered { background-color: pink  ; padding:0.7ex; margin:0.3ex; }

tr.IndexSummary td.indexSummary {
  background-color: #B0C4DE;
  border-color: white;
  border-style: double;
}
td.SumBarUncovered {
  background-color: Red;
}
td.SumBarCovered {
  background-color: Lime;
}
"

let draw_visual_bar co percentage =
  let percentage_uncov = 100 - percentage in
    output_string co ("<table border=\"0\" cellspacing=\"0\">");
    output_string co ("<tr height=\"10\">");
    if percentage <> 0 then
      output_string co ("<td class=SumBarCovered width=\""^string_of_int percentage^"\"></td>");
    if percentage_uncov <> 0 then
      output_string co ("<td class=SumBarUncovered width=\""^string_of_int percentage_uncov^"\"></td>");
    output_string co ("</tr></table>")

let is_complex_decision model =
  (* Was: List.length model.TraceMcdc.mod_conds > 1 *)
  match model.TraceMcdc.mod_conds with _::_::_ -> true | _ -> false

let string_of_file src =
  (* takes the whole contents of a file and puts it in a string *)
  (* N.B. we won't read from special files such as stdin anyway, 
     because this module is about making a report on a source code file that
     has been compiled and tested. *)
  let ic = open_in src in
  let b = Buffer.create (in_channel_length ic) in
    try
      while true do
        Buffer.add_char b (input_char ic)
      done;
      assert(false)
    with End_of_file -> Buffer.contents b



let build_decorated_source_code (src:string) (statements:Coverage.coverage_expr list) : decorated_char array =
  let covered_or_not = function 
    | Coverage.Covered   -> Covered
    | Coverage.Uncovered -> Notcovered
  in
  let sort_of_covev_status i = function
    | Covevent.Coverage_event_statement -> Statement i
    | Covevent.Coverage_event_decision -> Decision i
    | Covevent.Coverage_event_condition _ -> Condition i
  in
  let compare_statements = fun
    {Coverage.cnum_start=s;Coverage.cnum_end=e;_} {Coverage.cnum_start=s2;Coverage.cnum_end=e2;_} ->
      compare (s,e2) (s2,e)
  in
  try begin
    if Sys.file_exists src then 
      begin (********)
        let srcfile = string_of_file src in
        let sorted_statements = (List.sort compare_statements statements) in
          (* N.B. statements have to be sorted in this algorithm, else it doesn't work at all! *)
        Array.init (String.length srcfile) 
          (fun i ->
             let rec filter accu i = function
               | [] -> List.rev accu
               | { Coverage.event =
                     { Covevent.covev_status =
                         (Covevent.Coverage_event_statement
                         |Covevent.Coverage_event_decision
                         |Covevent.Coverage_event_condition _)
                           as covev_status;
                       _ };
                   Coverage.cnum_start = startpos;
                   Coverage.cnum_end = endpos;
                   Coverage.status = 
                     ((Coverage.Covered|Coverage.Uncovered) 
                        as coverage_status);
                 } :: next_statements ->
                   if startpos > i then
                     List.rev accu
                   else
                     begin
                       if startpos <= i && endpos >= i then
                         filter
                           ({ start_position = startpos;
                              end_position = endpos; 
                              expression_sort = sort_of_covev_status startpos covev_status; 
                              coverage_status = covered_or_not coverage_status;
                            } :: accu) 
                           i next_statements
                       else
                         filter accu i next_statements
                     end
             in (* / end filter *)
               { position = i,i; data = String.make 1 srcfile.[i] ; parents = filter [] i sorted_statements }
          )
      end   (********)
    else 
      failwith (Printf.sprintf  "source file <%s> has disappeared!" src)
  end with Sys_error _ -> failwith ("source file error html " ^ src)


let html_of_decorated_char_array (src:string) dca = 
  (* src: filename of the source code being processed *)
  let rec factorize =  function
    | ({ position = (a,b) ; data = data1 ; parents = parents1 } as first)
      :: ({ position = (x,y) ; data = data2 ; parents = parents2 } as second)
      :: next ->
        if parents1 = parents2 then 
          factorize ({ position = (min a x, max b y) ; data = data1 ^ data2 ; parents = parents1 }::next)
        else
          first :: factorize (second::next)
    | x ->
        x
  in
  let rec stringlist_of_parents = function
    | {start_position = start_position;
       end_position  = end_position;
       expression_sort = expression_sort;
       coverage_status = coverage_status } :: tl ->
        Printf.sprintf " (%d-%d:%s:%s) "
          start_position end_position
          (match expression_sort with 
            | Statement _ -> "statement" 
            | Condition _ -> "condition"  
            | Decision  _ ->  "decision")
          (match coverage_status with Covered -> "covered" | _ -> "notcovered")
        ^ stringlist_of_parents tl
    | [] -> ""
  in
  let determine_class_from_parents parents =
    let to_string = function 
      | { expression_sort = Decision _ ; coverage_status = Covered } ->
          "p-covered_decision"
      | { expression_sort = Decision _ ; coverage_status = Notcovered } ->
          "p-not_covered_decision"
      | { expression_sort = Condition _ ; coverage_status = Covered } ->
          "p-covered_condition"
      | { expression_sort = Condition _ ; coverage_status = Notcovered } ->
          "p-not_covered_condition"
      | { expression_sort = Statement _ ; coverage_status = Covered } ->
          "p-covered_statement"
      | { expression_sort = Statement _ ; coverage_status = Notcovered } ->
          "p-not_covered_statement"
    in
      List.fold_left (fun r p -> to_string p ^ " " ^ r) "" parents
  in
  let segregate parents =
    (* separate covered parents from not covered parents, for html display *)
    List.fold_left 
      (fun (c, n) -> function
        | { coverage_status = Covered } as e -> (e::c, n)
        | e -> (c, e::n)
      )
      ([],[])
      parents
  in
  let rec get_decision_id = function
      (* returns the first decision's ID. 
         N.B. We assume that other decisions won't be far anyway. *)
    | [] -> -1
    | { expression_sort = Decision i } :: _ -> i
    | _ :: parents -> get_decision_id parents    
  in
  let dcl = factorize (Array.to_list dca) in
    (List.map
        (function
          | { position = position ; data = data ; parents = [] } ->
              DATA (data)
          | { position = (x,y) ; data = data ; parents = parents } ->
              let covered_parents, not_covered_parents = segregate parents in
              SPAN (["class", determine_class_from_parents covered_parents;],
                [A (
                  (match get_decision_id parents with
                    | -1 -> [] 
                    | i ->
                        ["href",
                        Printf.sprintf "%s.mcdc.html#%d" (Filename.basename src) i] )
                  @
                    ["id", Printf.sprintf "%d" x; 
                     "class", determine_class_from_parents not_covered_parents;
                     "title", Printf.sprintf "code-location=[%d-%d] parents=(%s)" x y (stringlist_of_parents parents);
                     "style", "";
                     "onmouseover", "this.style.fontWeight='bold'";
                     "onmouseout", "this.style.fontWeight='normal'";
                    ],[DATA (data)]
                )])
        )
        dcl)


let generate_html (with_mcdc:bool) (src:string) (statements:Coverage.coverage_expr list) : unit =
  try begin
    if Sys.file_exists src then begin
      let ci = open_in src in
      let co = open_out ("coverage_report/"^Filename.basename src^".html") in
      output_string co "<HTML>";
      head co "Coverage report";
      (* write BODY *)
      output_string co ("<BODY>");
      output_string co ("<h1>ZamCov Expression Coverage for "^src^"</h1>\n");
      if with_mcdc then
        output_string co ("<h2>MC/DC measurement: <a href=\""^Filename.basename src^".mcdc.html\">"^src^".mcdc</a></h2>\n\n");
      (* ************************************************* *)
      output_string co "<div>";
      output_string co "<pre style='float:left;width:3em;border:none;'>";
      (* ***************BEGIN*************** *)
      (* Here we print line numbers on the left of the page *)
      for i = 1 to (
        let ic = open_in src in
        let x = ref 0 in
          try while true do ignore(input_line ic);incr x done;assert false
          with End_of_file -> !x
      ) do Printf.fprintf co "%4d\n" i done;
      output_string co "</pre>";
      (* ***************END  *************** *)
      output_string co "
<div style='position:absolute;left:3em;'>
<pre>";
      (let dsc = (* the decorated source code *)
        html_of_decorated_char_array src (build_decorated_source_code src statements) in
         List.iter (output_string co) (List.map string_of_html dsc));
      output_string co "</pre>"; 
      output_string co "
<br/><hr/><br/>
<h2>Legend (color information)</h2>
<pre style='line-height: 250%;'>
<span class='p-covered_decision'>I'm a covered decision, I'm underlined in green.</span>     
<span class='p-not_covered_decision'>I'm a not-covered decision, I'm underlined in red.</span> 
<span class='p-covered_condition'>I'm a covered condition, I'm overlined in green.</span>    
<span class='p-not_covered_condition'>I'm a not-covered condition, I'm overlined in red.</span>
<span class='p-covered_statement'>I'm a covered statement, I'm highlighted in light green.</span>    
<span class='p-not_covered_statement'>I'm a not-covered statement, I'm highlighted in light red.</span>
I don't have red nor green, it means that I'm free from coverage measurement.
</pre>

</div></div>
";
      output_string co "</BODY></HTML>";
      close_in ci;
      close_out co
    end;
  end with Sys_error _ -> failwith ("source file error html " ^ src)

let string_of_status statement = match statement.Coverage.status with
  Coverage.Covered -> "p-covered_decision p-covered_statement"
| Coverage.Uncovered -> "p-not_covered_decision"


let generate_mcdc_table src co ci statements mcdc =
  let decision_start statement =
    if statement.Coverage.event = mcdc.TraceMcdc.model.TraceMcdc.mod_dec then
      let id = string_of_int statement.Coverage.event.Covevent.covev_loc.Location.loc_start.Lexing.pos_cnum in
      output_string co
        ("<A NAME='" ^ id 
          ^ "' HREF='" ^ Filename.basename src ^ ".html#" ^ id ^
         "' CLASS=" ^ string_of_status statement ^ ">") in
  let print_condition cond =
    let l, l' = CoverUtils.get_loc cond in
    let cnum_start = l.Lexing.pos_cnum in
    let cnum_end = l'.Lexing.pos_cnum in
    let expr = String.create (cnum_end - cnum_start) in
    seek_in ci cnum_start;
    really_input ci expr 0 (cnum_end - cnum_start);
    output_string co "<TH>";
    output_string co (htmlspecialchars expr);
    output_string co "</TH>" in
  let print_cond_in_vec vector cond =
    output_string co "<TH>";
    for i=0 to Array.length vector.TraceMcdc.vec_conds - 1 do
      if cond = vector.TraceMcdc.vec_conds.(i).TraceMcdc.cond then begin
        match  vector.TraceMcdc.vec_conds.(i).TraceMcdc.cond_value with
          TraceMcdc.Bool true -> output_string co " T "
        | TraceMcdc.Bool false -> output_string co " F "
        | TraceMcdc.Unknown ->  output_string co " - "
      end
    done;
    output_string co "</TH>" in
  let print_vector vector =
    output_string co "<TR>";
    output_string co "<TH>";
    begin match vector.TraceMcdc.vec_dec.TraceMcdc.dec_value with
      TraceMcdc.Bool true -> output_string co " T "
    | TraceMcdc.Bool false -> output_string co " F "
    | TraceMcdc.Unknown ->  output_string co " - "
    end;
    let nb = string_of_int vector.TraceMcdc.vec_dec.TraceMcdc.nb_dec in
    output_string co (": "^nb);
    output_string co "</TH>";
    List.iter (print_cond_in_vec vector) mcdc.TraceMcdc.model.TraceMcdc.mod_conds;
    output_string co "</TR>" in
  output_string co "<TABLE BORDER=\"1\">";
  output_string co "<TR>";
  output_string co "<TH>";
  output_string co " <!-- decision --> ";
  List.iter decision_start statements;
  let l, l' = CoverUtils.get_loc mcdc.TraceMcdc.model.TraceMcdc.mod_dec in
  let line_start = l.Lexing.pos_lnum in
  output_string co ("line "^(string_of_int line_start)^": ");
  let cnum_start = l.Lexing.pos_cnum in
  let cnum_end = l'.Lexing.pos_cnum in
  let expr = String.create (cnum_end - cnum_start) in
  seek_in ci cnum_start;
  really_input ci expr 0 (cnum_end - cnum_start);
  output_string co (htmlspecialchars expr);
  output_string co "</A>";
  output_string co "</TH>";
  List.iter print_condition mcdc.TraceMcdc.model.TraceMcdc.mod_conds;
  List.iter print_vector mcdc.TraceMcdc.vectors;
  output_string co "</TR>";
  output_string co "</TABLE>";
  output_string co "<BR>"

let generate_mcdc_tabular src (_, mcdc_trace) statements =
  try begin
    if Sys.file_exists src then begin
      let ci = open_in src in
      let co = open_out ("coverage_report/"^Filename.basename src^".mcdc.html") in
      output_string co "<HTML>";
      head co "MC/DC report";
      (* write BODY *)
      output_string co ("<BODY>");
      output_string co ("<h3>ZamCov: MC/DC tabulars <br>("^src^")</h3><pre>\n");
      let filtered_trace =
        List.filter
          (fun mcdc -> is_complex_decision mcdc.TraceMcdc.model)
          mcdc_trace in
      List.iter (generate_mcdc_table src co ci statements) filtered_trace;
      output_string co "</HTML>";
      close_in ci;
      close_out co
    end
  end with Sys_error _ -> failwith ("source file error html" ^ src)

let generate_html_index exec all_cover_res exec_trace_file_names mcdc_trace_file_names =
  let co = open_out ("coverage_report/index.html") in

    head co ("Code coverage report for :"^exec);
    output_string co "<BODY style='padding:1em;'>\n";
    output_string co "<h2>Coverage report</h2>\n";

    (***************************************************)
    (* write the list of trace files used for coverage *)
    (***************************************************)
    output_string co "<hr/>\n";
    output_string co "<table CELLPADDING=1 class=\"IndexSummary\">";
    output_string co "
<tr class=\"IndexSummary\">
  <td class=\"IndexSummary\" align=center><b>Program Name</b></td>
</tr>\n";
    output_string co "<tr class=\"IndexSummary\">\n";
    output_string co ("<td class=\"IndexSummary\">"^exec^"</td></tr>\n");
    output_string co "</table>";

    output_string co "<table CELLPADDING=1 class=\"IndexSummary\">";
 output_string co "
<tr class=\"IndexSummary\">
  <td class=\"IndexSummary\" align=center><b>Trace Filename</b></td>
</tr>\n";
    List.iter
      (fun trace ->
	 output_string co "<tr class=\"IndexSummary\">\n";
	 output_string co ("<td class=\"IndexSummary\">"^trace^"</td></tr>\n"))
      exec_trace_file_names;
    output_string co "</table>";

    output_string co "<table CELLPADDING=1 class=\"IndexSummary\">";
 output_string co "
<tr class=\"IndexSummary\">
  <td class=\"IndexSummary\" align=center><b>MC/DC trace Filename</b></td>
</tr>\n";
    List.iter
      (fun trace ->
	 output_string co "<tr class=\"IndexSummary\">\n";
	 output_string co ("<td class=\"IndexSummary\">"^trace^"</td></tr>\n"))
      mcdc_trace_file_names;
    output_string co "</table>";

    (******************************************)
    (* write the list of covered source files *)
    (******************************************)
    output_string co "<hr/>\n";
    output_string co "<table CELLPADDING=1 class=\"IndexSummary\">";
    output_string co "
<tr class=\"IndexSummary\">
  <td class=\"IndexSummary\" align=center>Source Filename</td>
  <td class=\"IndexSummary\" align=center>Expression coverage</td>
</tr>
";
    List.iter
      (fun x ->
	 let src = x.Coverage.src in
	 let nb_events = x.Coverage.nb_events in
	 let nb_covered = nb_events - x.Coverage.nb_uncovered_events in
	 let percentage = int_of_float ((float nb_covered) /. (float nb_events) *. 100.) in
	   if Sys.file_exists src then 
	     begin
	       output_string co "<tr class=\"IndexSummary\"><td class=\"IndexSummary\">\n";
	       output_string co ("<a href=\""^Filename.basename src^".html\">\n");
	       output_string co (src^"</a></td>\n");
	       output_string co ("<td class=\"IndexSummary\" align=right>"^(string_of_int percentage)^"%  ");
	       output_string co ((string_of_int nb_covered)^"/ "^(string_of_int nb_events));
	       draw_visual_bar co percentage;
	     end)
      all_cover_res;
    output_string co "</table>\n";
    output_string co "<h2>Object code coverage report</h2>\n";
    output_string co ("<a href="^Filename.basename exec^".asm.html>"^exec^"</a>\n");
    output_string co "</BODY>";
    close_out co

let generate_html_asm exec data cov_status =
  let code = data.Bytecode_loader.code_section in
    try
      let co = open_out ("coverage_report/"^Filename.basename exec^".asm.html") in
	output_string co "<html>";
	head_object co;
	output_string co "<body><div class='bytecode'>";
	let aux cl inst i =
          output_string co (Printf.sprintf "%08d <span class='%s'>" i cl);
	  output_string co (Instructions.string_of_instructions data.Bytecode_loader.primitive_section inst);
	  output_string co "</span><br/>";
	in
	  for i=0 to Array.length code - 1 do
            match code.(i),cov_status.(i) with
              | Instructions.Param _, _ -> ()
              | inst, false -> aux "uncovered" inst i
              | inst, true -> aux "covered" inst i
	  done;
	  output_string co "</div>";
	  output_string co "</body>";
	  output_string co "</html>";
	  close_out co
    with
      | Sys_error _ -> failwith ("source file error obj " ^ exec)
